"""
API for using dbsamizdat as a library in Django
"""

from argparse import Namespace as _Namespace
from typing import Iterable, Union

from .runner import cmd_refresh as _cmd_refresh, cmd_sync as _cmd_sync, cmd_nuke as _cmd_nuke
from .const import txstyle, env
from .samizdat import Samizdat
from .exceptions import WrongTransactionStyleForParallelRefresh

_CMD_ARG_DEFAULTS = dict(
    context=env.API | env.DJANGO,
    verbosity=1,
    parallel=False,
)


def refresh(
    dbconn: str = 'default',
    transaction_style: txstyle = txstyle.CHECKPOINT,
    belownodes: Iterable[Union[str, tuple, Samizdat]] = tuple(),
    samizdatmodules=tuple(),
    parallel=False,
):
    """
    Refresh materialized views, in dependency order, optionally parallelized (where the dependency tree allows), optionally restricted to views depending directly or transitively on any of the DB objects specified in `belownodes`.
    """
    if parallel and transaction_style != txstyle.CHECKPOINT:
        raise WrongTransactionStyleForParallelRefresh()
    args = _Namespace(
        **{
            **_CMD_ARG_DEFAULTS,
            'dbconn': dbconn,
            'txdiscipline': transaction_style.value,
            'belownodes': belownodes,
            'samizdatmodules': [(None, m) for m in samizdatmodules],
            'parallel': parallel,
        }
    )
    _cmd_refresh(args)


def sync(
    dbconn: str = 'default',
    transaction_style: txstyle = txstyle.JUMBO,
    samizdatmodules=tuple(),
    parallel=False,
):
    """
    Sync dbsamizdat state to the DB.
    """
    if parallel and transaction_style != txstyle.CHECKPOINT:
        raise WrongTransactionStyleForParallelRefresh()
    args = _Namespace(
        **{
            **_CMD_ARG_DEFAULTS,
            'dbconn': dbconn,
            'txdiscipline': transaction_style.value,
            'samizdatmodules': [(None, m) for m in samizdatmodules],
            'parallel': parallel,
        }
    )
    _cmd_sync(args)


def nuke(
    dbconn: str = 'default',
    transaction_style: txstyle = txstyle.JUMBO,
    samizdatmodules=tuple(),
):
    """
    Remove any database object tagged as samizdat.
    """
    args = _Namespace(
        **_CMD_ARG_DEFAULTS,
        dbconn=dbconn,
        txdiscipline=transaction_style.value,
        samizdatmodules=[(None, m) for m in samizdatmodules],
    )
    _cmd_nuke(args)



import argparse
import logging
import sys
from time import monotonic
from concurrent.futures import ThreadPoolExecutor, as_completed
from os import cpu_count
from typing import Generator
from functools import partial
from pathlib import Path

try:
    try:
        import psycopg as psycopiggy
    except ImportError:
        import psycopg2 as psycopiggy
except ImportError as nope:
    raise RuntimeError("No psycopg or psycopg2 module available for import") from nope

from .libdb import dbinfo_to_class, dbstate_equals_definedstate, get_dbstate, DBObjectType
from .libgraph import depsort_with_sidekicks, node_dump, sanity_check, subtree_depends, deps_on_closure
from .graphvizdot import dot
from .loader import get_samizdats
from . import entitypes
from .const import txstyle, env
from .exceptions import SamizdatException, DatabaseError, FunctionSignatureError, WrongTransactionStyleForParallelRefresh
from .util import fqify_node, nodenamefmt, sqlfmt, honest_cursor


logger = logging.getLogger(__name__)
PRINTKWARGS = dict(file=sys.stderr, flush=True)
MAX_ACTION_VERB_LENGTH = 14
ENTITYTYPE_LENGTH = 17


def log(args: argparse.Namespace, *pargs, level=logging.INFO, **pkwargs):
    if (args.context & env.API):
        logger.log(level, ' '.join(map(str, pargs)))
    elif level > logging.INFO or (level, args.verbosity) in {(logging.INFO, 1), (logging.DEBUG, 2)}:
        print(*pargs, **{**PRINTKWARGS, **pkwargs})


def vprint(args, *pargs, **pkwargs):
    log(args, *pargs, **pkwargs)


def vvprint(args, *pargs, **pkwargs):
    log(args, *pargs, level=logging.DEBUG, **pkwargs)


def get_executor_drainer(args, max_namelen=0, parallel=False):
    if (args.context & env.API):
        return log_drain_executor
    return partial(chronoprint_drain_executor, args, max_namelen=max_namelen, parallel=parallel)


def log_drain_executor(executor: Generator):
    for timing, action_totake, sd, sql in executor:
        what = "STARTING" if timing is None else "FINISHED (%.2fs)" % timing
        logger.info(f"""{what} {action_totake} {sd.entity_type.value} {sd}""")
        logger.debug(f"""Generated SQL for {action_totake} {sd.entity_type.value} {sd}:\n{sqlfmt(sql)}""")


def chronoprint_drain_executor(args, executor: Generator, max_namelen=0, parallel=False):
    if parallel:
        # starts and finishes are interwoven
        for timing, action_totake, sd, sql in executor:
            what = "start " if timing is None else "finish"
            if timing is None:
                vprint(args, f'%-{MAX_ACTION_VERB_LENGTH}s %-{ENTITYTYPE_LENGTH}s %-{max_namelen}s' % (f'{what} {action_totake}', sd.entity_type.value, sd))
                vvprint(args, f'\n\n{sqlfmt(sql)}\n\n')
            else:
                vprint(args, f'%-{MAX_ACTION_VERB_LENGTH}s %-{ENTITYTYPE_LENGTH}s %-{max_namelen}s %7.2fs' % (f'{what} {action_totake}', sd.entity_type.value, sd, timing))
    else:
        # finishes directly follow starts for the same object
        for timing, action_totake, sd, sql in executor:
            if timing:
                vprint(args, '%7.2fs' % timing)
            else:
                vprint(args, f'%-{MAX_ACTION_VERB_LENGTH}s %-{ENTITYTYPE_LENGTH}s %-{max_namelen}s ' % (action_totake, sd.entity_type.value, sd), end='')
                vvprint(args, f'\n\n{sqlfmt(sql)}\n\n')


def get_cursor(args):
    cursor = None
    if (args.context & env.DJANGO):
        from django.db import connections
        cursor = honest_cursor(connections[args.dbconn].cursor())
    else:
        cursor = psycopiggy.connect(args.dburl).cursor()
    cursor.execute('BEGIN;')  # And so it begins…
    return cursor


def txi_finalize(cursor, args):
    do_what = {txstyle.JUMBO.value: 'COMMIT;', txstyle.DRYRUN.value: 'ROLLBACK;'}.get(args.txdiscipline)
    if do_what:
        cursor.execute(do_what)


def cmd_refresh(args):
    if args.parallel and args.txdiscipline != txstyle.CHECKPOINT.value:
        raise WrongTransactionStyleForParallelRefresh()

    cursor = get_cursor(args)
    samizdats = depsort_with_sidekicks(sanity_check(get_samizdats(args.samizdatmodules, context=args.context)))
    matviews = [sd for sd in samizdats if sd.entity_type == entitypes.MATVIEW]

    if args.belownodes:
        rootnodes = {fqify_node(rootnode) for rootnode in args.belownodes}
        allnodes = node_dump(samizdats)
        if rootnodes - allnodes:
            raise ValueError('''Unknown rootnodes:\n\t- %s''' % '\n\t- '.join([nodenamefmt(rootnode) for rootnode in rootnodes - allnodes]))
        subtree_bundle = subtree_depends(samizdats, rootnodes)
        matviews = [sd for sd in matviews if sd in subtree_bundle]

    max_namelen = max(len(str(ds)) for ds in matviews)

    def refreshes(subjects):
        for sd in subjects:
            yield 'refresh', sd, sd.refresh(concurrent_allowed=True)

    drain = get_executor_drainer(args, max_namelen=max_namelen, parallel=args.parallel)
    if args.parallel:
        firstlevel, parallelization_callback = prep_refresh_parallel(samizdats, matviews, refreshes)
        drain(executor(refreshes(firstlevel), args, None, parallelization_callback=parallelization_callback))
    else:
        drain(executor(refreshes(matviews), args, cursor))

    txi_finalize(cursor, args)


def prep_refresh_parallel(samizdats, matviews, refreshes):
    deps_closure = {
        sd: {d for d in deps if d in matviews}
        for sd, deps in deps_on_closure(samizdats).items()
        if sd in matviews
    }

    depfree = {sd for sd, deps in deps_closure.items() if not deps}
    done = set()
    submitted = set(depfree)

    def paralellization_callback(sd):
        nonlocal submitted
        done.add(sd)
        left_todo = (deps_closure.keys() - done) - submitted
        if not left_todo:
            return False  # signal that we're done
        submittables = {sd for sd in left_todo if deps_closure[sd].issubset(done)}
        submitted |= submittables
        return refreshes(submittables)

    return depfree, paralellization_callback


def cmd_sync(args):
    if args.parallel and args.txdiscipline != txstyle.CHECKPOINT.value:
        raise WrongTransactionStyleForParallelRefresh()

    cursor = get_cursor(args)
    samizdats = depsort_with_sidekicks(sanity_check(get_samizdats(args.samizdatmodules, context=args.context)))
    issame, excess_dbstate, excess_definedstate = dbstate_equals_definedstate(cursor, samizdats)
    if issame:
        vprint(args, "No differences, nothing to do.")
        return
    max_namelen = max(len(str(ds)) for ds in excess_dbstate | excess_definedstate)
    drain = get_executor_drainer(args, max_namelen=max_namelen)
    if excess_dbstate:
        def drops():
            for sd in excess_dbstate:
                yield 'drop', sd, sd.drop(if_exists=True)  # we don't know the deptree; so they may have vanished through a cascading drop of a previous object
        drain(executor(drops(), args, cursor))
        issame, excess_dbstate, excess_definedstate = dbstate_equals_definedstate(cursor, samizdats)  # again, we don't know the in-db deptree, so we need to re-read DB state as the rug may have been pulled out from under us with cascading drops
    if excess_definedstate:
        matviews_to_refresh = []
        def creates():
            to_create_ids = {sd.head_id() for sd in excess_definedstate}
            for sd in samizdats:  # iterate in proper creation order
                if sd.head_id() in to_create_ids:
                    yield 'create', sd, sd.create()
                    yield 'sign', sd, sd.sign()
                    if sd.entity_type == entitypes.MATVIEW:
                        matviews_to_refresh.append(sd)  # queue to populate this matview later
        drain(executor(creates(), args, cursor))

        def refreshes(subjects):
            for sd in subjects:
                yield 'refresh', sd, sd.refresh(concurrent_allowed=False)

        if matviews_to_refresh:
            drain = get_executor_drainer(args, max_namelen=max_namelen, parallel=args.parallel)
            if args.parallel:
                firstlevel, parallelization_callback = prep_refresh_parallel(samizdats, matviews_to_refresh, refreshes)
                drain(executor(refreshes(firstlevel), args, None, parallelization_callback=parallelization_callback))
            else:
                drain(executor(refreshes(matviews_to_refresh), args, cursor))

    txi_finalize(cursor, args)


def cmd_diff(args):
    cursor = get_cursor(args)
    samizdats = depsort_with_sidekicks(sanity_check(get_samizdats(args.samizdatmodules, context=args.context)))
    issame, excess_dbstate, excess_definedstate = dbstate_equals_definedstate(cursor, samizdats)
    if issame:
        vprint(args, "No differences.")
        exit(0)

    max_namelen = max(len(str(ds)) for ds in excess_dbstate | excess_definedstate)

    def statefmt(state, prefix):
        return '\n'.join(f'%s%-17s\t%-{max_namelen}s\t%s' % (prefix, sd.entity_type.value, sd, sd.definition_hash()) for sd in sorted(state, key=lambda sd: str(sd)))
    if excess_dbstate:
        vprint(args, statefmt(excess_dbstate, 'Not in samizdats:\t'), file=sys.stdout)
    if excess_definedstate:
        vprint(args, statefmt(excess_definedstate, 'Not in database:   \t'), file=sys.stdout)
    exit(100 + (1 if excess_dbstate else 0 | 2 if excess_definedstate else 0))


def cmd_printdot(args):
    print('\n'.join(dot(depsort_with_sidekicks(sanity_check(get_samizdats(args.samizdatmodules, context=args.context))))))


def cmd_nuke(args, samizdats=None):
    cursor = get_cursor(args)

    def nukes():
        nonlocal samizdats
        if samizdats is None:
            samizdats = map(dbinfo_to_class, filter(lambda a: a[-1] is not None, get_dbstate(cursor)))
        for sd in samizdats:
            yield ("nuke", sd, sd.drop(if_exists=True))

    max_namelen = max(len(str(ds)) for ds in samizdats) if samizdats else 0
    drain = get_executor_drainer(args, max_namelen=max_namelen)

    drain(executor(nukes(), args, cursor))
    txi_finalize(cursor, args)


def executor(plan, args, cursor, parallelization_callback=None):

    def actually_exec(action_totake, sd, sql):
        local_cursor = get_cursor(args) if parallelization_callback else cursor  # For parallel refreshes, get a cursor on new connection (we'll get a new connection, as this is a new thread)
        starttime = monotonic()
        try:
            try:
                local_cursor.execute("BEGIN;")  # harmless if already in a tx
                local_cursor.execute(f"SAVEPOINT action_{action_totake};")
                local_cursor.execute(sql)
            except psycopiggy.errors.UndefinedFunction as ouch:
                if action_totake == 'sign':
                    local_cursor.execute(f"ROLLBACK TO SAVEPOINT action_{action_totake};")  # get back to a non-error state
                    candidate_args = [c[3] for c in get_dbstate(local_cursor, which=DBObjectType.FOREIGN, entity_types=(entitypes.FUNCTION,)) if c[:2] == (sd.schema, sd.function_name)]
                    raise FunctionSignatureError(sd, candidate_args)
                raise ouch
        except psycopiggy.Error as dberr:
            raise DatabaseError(f"{action_totake} failed", dberr, sd, sql)
        local_cursor.execute(f'RELEASE SAVEPOINT action_{action_totake};')
        if args.txdiscipline == txstyle.CHECKPOINT.value and action_totake != 'create':
            # only commit *after* signing, otherwise if later the signing somehow fails we'll have created an orphan DB object that we don't recognize as ours
            local_cursor.execute("COMMIT;")
        return (monotonic() - starttime, action_totake, sd, sql)

    if parallelization_callback:
        pool = ThreadPoolExecutor(cpu_count(), thread_name_prefix="dbsamizdat-refresh-")
        tasks = []
        # initial items
        for planitem in plan:
            yield (None, *planitem)
            tasks.append(pool.submit(actually_exec, *planitem))

        done = False
        completed = set()
        while not done:
            for future in as_completed(reversed(tasks)):
                if future in completed:
                    continue
                yield future.result()
                completed.add(future)
                *_, sd, _sql = future.result()
                more_to_do = parallelization_callback(sd)
                if more_to_do is False:
                    done = True
                else:
                    for planitem in more_to_do:
                        yield (None, *planitem)
                        tasks.append(pool.submit(actually_exec, *planitem))
        pool.shutdown()

    else:
        for planitem in plan:
            yield (None, *planitem)
            runtime, *_ = actually_exec(*planitem)
            yield (runtime, *planitem)


def augment_argument_parser(p, context):

    def perhaps_add_modules_argument(parser):
        if not (context & env.DJANGO):
            parser.add_argument('samizdatmodules', nargs='+', help='Module paths or python files containing Samizdat subclasses. Use package/module paths such as "somepackage.mymodule" or filepaths such as "../path/to/mymodule.py".')

    def add_dbarg_argument(parser):
        if (context & env.DJANGO):
            parser.add_argument('dbconn', nargs='?', default='default', help="Django DB connection key (default:'default'). If you don't know what this is, then you don't need it.")
        else:
            parser.add_argument('dburl', help="PostgreSQL DB connection string. Trivially, this might be 'postgresql:///mydbname'. See https://www.postgresql.org/docs/14/static/libpq-connect.html#id-1.7.3.8.3.6 .")

    def add_txdiscipline_argument(parser):
        parser.add_argument('--txdiscipline', '-t', choices=(txstyle.CHECKPOINT.value, txstyle.JUMBO.value, txstyle.DRYRUN.value), default=txstyle.CHECKPOINT.value, help=f"""Transaction discipline. The default "{txstyle.CHECKPOINT.value}" level commits after every dbsamizdat-level action, and is compatible with parallelized materialized view refreshing. "{txstyle.JUMBO.value}" creates one large transaction. "{txstyle.DRYRUN.value}" also creates one large transaction, but rolls it back.""")

    def add_parallel_argument(parser):
        parser.add_argument('--parallel', '-p', action="store_true", help=f'Parallelize refreshing of materialized views. Can only be used with the "{txstyle.CHECKPOINT.value}" transaction discipline.')

    p.set_defaults(
        func=lambda whatevs: p.print_help(),
        context=context,
        samizdatmodules=[],
        verbosity=1,
        parallel=False,
    )
    if not (context & env.DJANGO):
        p.add_argument('--quiet', '-q', help="Be quiet (minimal output)", action="store_const", const=0, dest='verbosity')
        p.add_argument('--verbose', '-v', help="Be verbose (on stderr).", action="store_const", const=2, dest='verbosity')
    else:
        p.add_argument('-v', '--verbosity', default=1, type=int)
    subparsers = p.add_subparsers(title='commands')

    p_nuke = subparsers.add_parser('nuke', help='Drop all dbsamizdat database objects.')
    p_nuke.set_defaults(func=cmd_nuke)
    add_txdiscipline_argument(p_nuke)
    add_dbarg_argument(p_nuke)

    p_printdot = subparsers.add_parser('printdot', help='Print DB object dependency tree in GraphViz format.')
    p_printdot.set_defaults(func=cmd_printdot)
    perhaps_add_modules_argument(p_printdot)

    p_diff = subparsers.add_parser('diff', help='Show differences between dbsamizdat state and database state. Exits nonzero if any are found: 101 when there are excess DB-side objects, 102 if there are excess python-side objects, 103 if both sides have excess objects.')
    p_diff.set_defaults(func=cmd_diff)
    add_dbarg_argument(p_diff)
    perhaps_add_modules_argument(p_diff)

    p_refresh = subparsers.add_parser('refresh', help='Refresh materialized views, in dependency order.')
    p_refresh.set_defaults(func=cmd_refresh)
    add_txdiscipline_argument(p_refresh)
    add_parallel_argument(p_refresh)
    add_dbarg_argument(p_refresh)
    perhaps_add_modules_argument(p_refresh)
    p_refresh.add_argument('--belownodes', '-b', nargs='*', help="Limit to views that depend on ENTITYNAMES (usually, specific tables).", metavar='ENTITYNAMES')

    p_sync = subparsers.add_parser('sync', help='Make it so!')
    p_sync.set_defaults(func=cmd_sync)
    add_txdiscipline_argument(p_sync)
    add_parallel_argument(p_sync)
    add_dbarg_argument(p_sync)
    perhaps_add_modules_argument(p_sync)


def convert_paths_to_modules(args):

    def perhaps_filename_to_module(thing: str):
        if thing.endswith('.py'):
            if (maybe_file:= Path(thing).expanduser()).is_file():

                maybe_module_name = maybe_file.stem
                if '.' in maybe_module_name:
                    exit(f'Fatal: Cannot load "{thing}" as a module: filename contains a period (".")')

                if (maybe_shadowing_dir := maybe_file.parent / maybe_module_name).is_dir():
                    exit(f'Fatal: Cannot load "{thing}" as module: directory "{maybe_shadowing_dir}" shadows module name "{maybe_module_name}"')

                dir_containing_module = maybe_file.parent.resolve()
                return (dir_containing_module, maybe_module_name)
            else:
                exit(f'Fatal: Cannot load "{thing}": not a Python file')
        return (None, thing)

    args.samizdatmodules = [perhaps_filename_to_module(thing) for thing in args.samizdatmodules]

def main():
    p = argparse.ArgumentParser(description='dbsamizdat, the blissfully naive PostgreSQL database object manager.')
    augment_argument_parser(p, context=env.CLI)
    args = p.parse_args()
    convert_paths_to_modules(args)
    try:
        args.func(args)
    except SamizdatException as argh:
        exit(f'\n\n\nFATAL: {argh}')
    except KeyboardInterrupt:
        exit('\nInterrupted.')


if __name__ == '__main__':
    main()



from enum import Enum, IntFlag, auto

class txstyle(Enum):
    CHECKPOINT = 'checkpoint'
    JUMBO = 'jumbo'
    DRYRUN = 'dryrun'

class env(IntFlag):
    API = auto()
    CLI = auto()
    DJANGO = auto()



from logging import getLogger
from importlib import import_module
import inspect
from itertools import chain
import sys

from .samizdat import SamizdatView, SamizdatMaterializedView, SamizdatFunction, SamizdatTrigger, SamizdatMeta, SamizdatFunctionMeta, SamizdatTriggerMeta
from .const import env

logger = getLogger(__name__)

AUTOLOAD_MODULENAME = "dbsamizdat_defs"

def module_not_found_help(modulename, exception, pypath):
    withsearchpath = f'''using Python import search path "{pypath}" ''' if pypath is not None else ''
    return f'''
Fatal: Loading module "{modulename}" {withsearchpath}failed with error:

=====================================================================
{exception}
=====================================================================

If you're unsure on how to solve this, try one of the following suggestions:
a.  Invoke dbsamizdat while you are inside a directory from which your module is importable.
    Thus `python -m {modulename}` should work from there, too.
b.  Set the PYTHONPATH environment variable to the directory from which your module is importable.
    Thus `PYTHONPATH="/path/to/some/project` python -m {modulename}` should work, too.
c.  If the error points to problems importing third-party modules, and you're using these
    third-party packages from a virtual environment, it'd be best to simply install dbsamizdat
    inside that virtual environment, too, and invoke your virtualenv's version of dbsamizdat.
    If that's not possible try using the PYTHONPATH environment variable to make those modules
    importable. You can see which paths you might want to add by loading your virtualenv and
    executing `python -c 'import sys; print("\\n".join(filter(None, sys.path)))'` .

    Documentation on the PYTHONPATH environment variable can be found here:
    https://docs.python.org/3/using/cmdline.html#envvar-PYTHONPATH
'''


def get_samizdats(modulelist=tuple(), context=0):

    def issamizdat(thing):
        excluded_classes = {SamizdatView, SamizdatMaterializedView, SamizdatFunction, SamizdatTrigger}
        return inspect.isclass(thing) and isinstance(thing, (SamizdatMeta, SamizdatFunctionMeta, SamizdatTriggerMeta)) and (thing not in excluded_classes)

    sdmodules = []
    if not (context & env.DJANGO):
        for (pypath, modulename) in modulelist:
            try:
                if pypath is None:
                    sdmodules.append(import_module(modulename))
                else:
                    sys.path.insert(0, str(pypath))
                    sdmodules.append(import_module(modulename))
                    sys.path.pop(0)
            except ModuleNotFoundError as notfounderror:
                if (context & env.CLI):
                    exit(module_not_found_help(modulename, notfounderror, pypath))
                else:
                    raise
    else:
        # if we're running in Django, we will autoload definitions from:
        # - the modules named in settings.DBSAMIZDAT_MODULES
        # - the module with name AUTOLOAD_MODULENAME of each app
        try:
            from django.core.exceptions import ImproperlyConfigured
            from django.conf import settings
            from django.apps import apps
        except ImportError as e:
            exit(f"Loading Django modules failed:\n{e}")
        else:
            try:
                django_sdmodules = [import_module(sdmod) for sdmod in getattr(settings, 'DBSAMIZDAT_MODULES', [])]
                for appconfig in apps.get_app_configs():
                    try:
                        django_sdmodules.append(import_module('{}.{}'.format(appconfig.module.__package__, AUTOLOAD_MODULENAME)))
                    except ModuleNotFoundError as err:
                        if not err.msg.endswith(f"{AUTOLOAD_MODULENAME}'"):
                            raise err
                if not django_sdmodules:
                    logger.warn(f"""No settings.DBSAMIZDAT_MODULES defined, and none of your apps contain any "{AUTOLOAD_MODULENAME}" module to autoload.""")
                sdmodules += django_sdmodules
            except ImproperlyConfigured:
                # assume we're not running in a fully booted Django
                pass

    return {c for cname, c in chain.from_iterable(map(lambda m: inspect.getmembers(m, issamizdat), sdmodules))}



from argparse import Namespace

from django.core.management.base import BaseCommand

from dbsamizdat.runner import augment_argument_parser
from dbsamizdat.exceptions import SamizdatException
from dbsamizdat.const import env


class Command(BaseCommand):
    help = 'Play nice with dbsamizdat.'

    def create_parser(self, *args, **kwargs):
        return super(Command, self).create_parser(*args, **{**kwargs, **{'conflict_handler': 'resolve'}})

    def add_arguments(self, parser):
        augment_argument_parser(parser, context=env.DJANGO)

    def handle(self, *args, **options):
        try:
            options['func'](Namespace(**{'parallel': False, **options}))
        except SamizdatException as argh:
            exit(f'\n\n\nFATAL: {argh}')
        except KeyboardInterrupt:
            exit('\nInterrupted.')
